from __future__ import annotations

import re

import numpy as np
import pandas as pd


def set_name(df, new_name):
    df = df.copy()
    old_name = df.loc[0, 'Name']

    df.loc[0, 'Name'] = new_name
    pattern = re.compile(r'^' + re.escape(old_name) + r'(?=\s*>>|$)', flags=re.IGNORECASE)
    df['Path'] = df['Path'].str.replace(pattern, new_name, regex=True)
    df['ID'] = np.nan

    return df


def upsert(df1, df2, prefer_right=True):
    """
    Upserts the data from df2 into df1 based on case-insensitive Path and Name values.
    If a row from df2 matches a row in df1, and the two have conflicting values, then preference
    is given as per the prefer_right parameter. Keeps the columns of df1
    """
    if len(df2) == 0:
        return df1
    if len(df1) == 0:
        return df2

    orig_columns = df1.columns
    df1 = df1.copy()
    df2 = df2.copy()
    for df in (df1, df2):
        df['path_nocase'] = df.Path.astype('object').str.casefold()
        df['name_nocase'] = df.Name.astype('object').str.casefold()
    df = df1.merge(df2, how='outer', on=['path_nocase', 'name_nocase'])
    wipe_ids = pd.Series(False, index=df.index)
    for column in orig_columns:
        prefer_right_column = prefer_right and column not in ['Path', 'Name']
        left_column = column + '_x'
        right_column = column + '_y'
        if right_column in df.columns:
            prefer_column = right_column if prefer_right_column else left_column
            backup_column = left_column if prefer_right_column else right_column
            df[column] = df[prefer_column]
            missing_values = pd.isnull(df[column])
            df.loc[missing_values, column] = df.loc[missing_values, backup_column]
            df[column] = df[column].apply(safe_int_cast)
            if column == 'Type' and 'ID' in df.columns:
                wipe_ids = wipe_ids | df.apply(lambda row: type_differs(row[prefer_column], row[backup_column]),
                                               axis=1)
    df.drop(columns=df.columns.difference(orig_columns), inplace=True)
    if 'ID' in df.columns:
        df.loc[wipe_ids, 'ID'] = np.nan
    return df


def drop_duplicate_items(df):
    """
    Removes duplicate items (identified by case-insensitive Path and Name) from a dataframe.
    """
    if len(df) == 0:
        return
    df['path_nocase'] = df.Path.astype('object').str.casefold()
    df['name_nocase'] = df.Name.astype('object').str.casefold()
    df.drop_duplicates(subset=['path_nocase', 'name_nocase'], inplace=True, ignore_index=True)
    df.drop(columns=['path_nocase', 'name_nocase'], inplace=True)


def safe_int_cast(x):
    return int(x) if isinstance(x, float) and not np.isnan(x) and x == int(x) else x


def initialize_status_df(status, action, *args):
    status.df = pd.DataFrame([{
        f'Assets {action}': 0,
        f'Signals {action}': 0,
        f'Conditions {action}': 0,
        f'Scalars {action}': 0,
        f'Metrics {action}': 0,
        f'Displays {action}': 0,
        f'Total Items {action}': 0,
        f'Items Pulled From Seeq': 0,
        f'Errors Encountered': 0
    }], index=['Status'])
    status.update(*args)


def increment_status_df(status, new_items=None, pulled_items=None, error_items=None, subtract_errors=False):
    if new_items is not None and len(new_items):
        for column in status.df.columns:
            if 'Type' in new_items.columns:
                for item_type in ['Asset', 'Signal', 'Condition', 'Scalar', 'Metric', 'Display']:
                    if item_type in column:
                        status.df[column] += sum(new_items['Type'].fillna('').str.contains(item_type))
            if 'Total' in column:
                status.df[column] += len(new_items)
    if pulled_items is not None:
        status.df['Items Pulled From Seeq'] += len(pulled_items)
    if error_items is not None and len(error_items):
        if subtract_errors:
            for column in status.df.columns:
                if 'Type' in error_items.columns:
                    for item_type in ['Asset', 'Signal', 'Condition', 'Scalar', 'Metric', 'Display']:
                        if item_type in column:
                            status.df[column] -= sum(error_items['Type'].fillna('').str.contains(item_type))
                if 'Total' in column:
                    status.df[column] -= len(error_items)
        status.df['Errors Encountered'] += len(error_items)
    status.update()


def type_differs(t1, t2):
    if pd.isnull(t1) or pd.isnull(t2) or len(t1) == 0 or len(t2) == 0:
        return False
    if 'Calculated' in t1 and ('Stored' in t2 or 'Literal' in t2):
        return False
    if ('Stored' in t1 or 'Literal' in t1) and 'Calculated' in t2:
        return False
    for simple_type in ('Asset', 'Scalar', 'Signal', 'Condition', 'Metric', 'Display'):
        if simple_type in t1 and simple_type in t2:
            return False
    return True


def visualize(df):
    if df.iloc[0].Depth != 1:
        df = df.copy()
        df['Depth'] = df['Depth'] - df.iloc[0].Depth + 1

    show_vertical_columns = pd.Series([False] * (df['Depth'].max() - 2), dtype=bool)
    lines = []
    for idx in reversed(df.index):
        depth = df.loc[idx, 'Depth']
        name = df.loc[idx, 'Name']
        if depth == 1:
            line = name
        else:
            line = ''
            for i in range(depth - 2):
                if show_vertical_columns[i]:
                    line += '|   '
                else:
                    line += '    '
            line += '|-- '
            line += name
            show_vertical_columns[depth - 2] = True
            show_vertical_columns[depth - 1:] = False
        lines.append(line)

    return '\n'.join(reversed(lines))



from __future__ import annotations

from typing import Optional, Union

import pandas as pd

from seeq.base import util
from seeq.spy import _common
from seeq.spy._errors import *
from seeq.spy._status import Status
from seeq.spy.workbooks.job.data import _pull, _push


def redo(job_folder: str, items_df: Union[pd.DataFrame, str, list], action: Optional[str] = None,
         *, quiet: Optional[bool] = None, status: Optional[Status] = None):
    """
    Creates a zip file of the job folder for easy sharing.

    Parameters
    ----------
    job_folder : {str}
        A full or partial path to the job folder to be zipped.

    items_df : {pd.DataFrame, str, list}
        A DataFrame containing an 'ID' column that can be used to identify
        the data items to affect. These IDs are based on the source system
        (not the destination system). Alternatively, you can supply an item
        ID directly as a str or list of strs.

    action : str
        If supplied, limits the redo to the specified actions. You can specify
        'pull' or 'push'. If not supplied, both pull and push are affected.
        Note that 'pull' automatically includes 'push'.

    quiet : bool
        If True, suppresses progress output. Note that when status is
        provided, the quiet setting of the Status object that is passed
        in takes precedence.

    status : spy.Status, optional
        If specified, the supplied Status object will be updated as the command
        progresses. It gets filled in with the same information you would see
        in Jupyter in the blue/green/red table below your code while the
        command is executed.
    """
    _common.validate_argument_types([
        (job_folder, 'job_folder', str),
        (items_df, 'items_df', (pd.DataFrame, str, list)),
        (action, 'action', str),
        (quiet, 'quiet', bool),
        (status, 'status', Status)
    ])

    status = Status.validate(status, quiet)

    if not util.safe_exists(job_folder):
        raise SPyValueError(f'Job folder "{job_folder}" does not exist.')

    if isinstance(items_df, str):
        items_df = pd.DataFrame({'ID': [items_df]})
    elif isinstance(items_df, list):
        items_df = pd.DataFrame({'ID': items_df})

    if 'ID' not in items_df:
        raise SPyValueError('items_df must contain an ID column.')

    if action is not None:
        if action not in ['pull', 'push']:
            raise SPyValueError('action must be "pull" or "push".')
    else:
        action = 'pull'

    items_df: pd.DataFrame
    status_columns = [c for c in ['ID', 'Name', 'Type'] if c in items_df]

    status.df = items_df[status_columns].copy()

    if action == 'pull':
        _pull.redo(job_folder, status)

    if action == 'push':
        _push.redo(job_folder, status)

    status.update(
        'Successfully marked specified items to be redone. Execute spy.job.workbooks.data.pull() and/or '
        'spy.job.workbooks.data.push() again.', Status.SUCCESS)



from __future__ import annotations

from typing import Dict, List, Optional, Union

import pandas as pd

from seeq import spy
from seeq.spy import _common
from seeq.spy._errors import *


class ItemMap:
    """
    Represents a map of item identifiers from a source (usually a saved set of workbooks) to the destination on the
    Seeq Server. This class is used extensively by spy.workbooks.push() operations.
    """
    _item_map: Union[dict, ItemMap]
    _lookup_df: Optional[pd.DataFrame]
    _logs: Dict[str, List[str]]
    _dummy_items: pd.DataFrame

    only_override_maps: bool
    data_item_cache: _common.LRUCache

    def __getstate__(self):
        return self._item_map, self._lookup_df, self._logs, self._dummy_items, self.only_override_maps

    def __setstate__(self, state):
        self.data_item_cache = _common.LRUCache()
        (self._item_map, self._lookup_df, self._logs, self._dummy_items, self.only_override_maps) = state

    def __init__(self, item_map=None, lookup_df: Optional[pd.DataFrame] = None):
        self._item_map = item_map if item_map is not None else dict()
        self._lookup_df = lookup_df
        self._logs = dict()
        self._dummy_items = pd.DataFrame()
        self.only_override_maps = False
        self.data_item_cache = _common.LRUCache()

    def __contains__(self, key):
        key = _common.ensure_upper_case_id('ID', key)
        return self._item_map.__contains__(key)

    def __getitem__(self, key):
        key = _common.ensure_upper_case_id('ID', key)
        try:
            return self._item_map.__getitem__(key)
        except KeyError:
            raise KeyError(f"Item {key} not found in item map. It is either not in the workbooks' inventory or "
                           f"the inventory is not being pushed as part of this operation.")

    def __setitem__(self, key, val):
        key = _common.ensure_upper_case_id('ID', key)
        val = _common.ensure_upper_case_id('ID', val)
        self._item_map.__setitem__(key, val)

    def __delitem__(self, key):
        key = _common.ensure_upper_case_id('ID', key)
        self._item_map.__delitem__(key)

    def keys(self):
        return self._item_map.keys()

    @property
    def has_look_up_df(self):
        return self._lookup_df is not None

    def look_up_id(self, value):
        row = _common.look_up_in_df(value, self._lookup_df)
        return row['ID']

    def log(self, key, message, at_top=False):
        if key not in self._logs:
            self._logs[key] = list()
        if at_top:
            self._logs[key].insert(0, message)
        else:
            self._logs[key].append(message)

    def add_dummy_item(self, dummy_item: pd.DataFrame):
        self._dummy_items = pd.concat([self._dummy_items, dummy_item])

    @property
    def dummy_items(self):
        return self._dummy_items

    def explain(self, key: Union[str, pd.DataFrame, pd.Series]):
        """
        Returns a string explaining the mapping outcome of a particular item.
        This is useful for debugging datasource mapping, both for things that
        fail and things that succeed to learn how the mapping was applied.
        Typically, you will use print() to show the returned value in the
        console.

        :param key: The item ID of the item to be explained. This can also be
        a one-row DataFrame or a Series with an 'ID' column. Note that this
        ID is of the item in the source workbook, not the ID of the pushed item
        in the destination workbook.
        :return: A string explaining the mapping outcome of a particular item.
        """
        if isinstance(key, pd.DataFrame):
            if 'ID' not in key.columns:
                raise SPyValueError('explain() requires a one-row DataFrame with an "ID" column')
            if len(key) == 0:
                raise SPyValueError('DataFrame supplied is empty')
            if len(key) > 1:
                raise SPyValueError('DataFrame supplied has more than one row')
            key = key.iloc[0]['ID']
        elif isinstance(key, pd.Series):
            if 'ID' not in key:
                raise SPyValueError('explain() requires a Series with "ID"')
            key = key['ID']
        elif not isinstance(key, str):
            raise SPyTypeError(f'Argument to explain() must be a str, DataFrame, or Series, '
                               f'not {key.__class__.__name__}')

        if key not in self._logs:
            return f'There are no mapping logs for {key}'

        return '\n'.join(self._logs[key])

    def clear_logs(self, key):
        self._logs[key] = list()


class OverrideItemMap(ItemMap):
    """
    Takes an existing ItemMap and overrides various keys. This is used extensively in the templating system to
    temporarily (for the extent of a particular frame in a callstack) override the inner ItemMap using
    template_parameters or an override_map.
    """

    _override: dict
    _parameters: Optional[dict]

    def __init__(self, item_map: ItemMap, *, template_parameters: dict = None, override_map: dict = None):
        super().__init__(item_map)

        self._override = dict() if override_map is None else override_map
        self._parameters = None
        if template_parameters:
            self._parameters = template_parameters
            self._override_from_template_parameters(template_parameters)

    def __contains__(self, key):
        key = _common.ensure_upper_case_id('ID', key)
        return self._override.__contains__(key) or self._item_map.__contains__(key)

    def __getitem__(self, key):
        key = _common.ensure_upper_case_id('ID', key)
        if self._override.__contains__(key):
            return self._override.__getitem__(key)
        else:
            return super().__getitem__(key)

    def __delitem__(self, key):
        key = _common.ensure_upper_case_id('ID', key)
        if self._override.__contains__(key):
            self._override.__delitem__(key)
        else:
            self._item_map.__delitem__(key)

    @property
    def has_look_up_df(self):
        return self._item_map.has_look_up_df

    def look_up_id(self, value):
        return self._item_map.look_up_id(value)

    def keys(self):
        keys = set(self._item_map.keys())
        keys.update(self._override.keys())
        return keys

    def override(self, key, value):
        self._override[key] = value

    @property
    def parameters(self):
        if self._parameters is not None:
            return self._parameters

        if isinstance(self._item_map, OverrideItemMap):
            return self._item_map.parameters

        return None

    def _override_from_template_parameters(self, template_parameters: dict):

        for key, value in template_parameters.items():
            if value is None or (value is float and pd.isna(value)):
                continue

            _id, _type, _fqn = spy.workbooks.ItemTemplate.code_key_tuple(key)

            if _id is None:
                # This is the case of a Mustachioed annotation with {{variable}} tokens in it
                continue

            if isinstance(value, pd.DataFrame):
                if len(value) > 1:
                    raise SPyValueError(f'Multiple rows in template_parameters dict for "{key}":\n{value}')
                if len(value) == 0:
                    raise SPyValueError(f'Empty DataFrame in template_parameters dict for "{key}"')
                value = value.iloc[0].to_dict()
            elif isinstance(value, str):
                value = {'ID': value} if _common.is_guid(value) else {'Name': value}
            elif isinstance(value, spy.workbooks.ItemTemplate):
                continue

            if _common.present(value, 'ID') and not _common.get(value, 'Reference', default=False):
                self._override[_id] = _common.get(value, 'ID')
            else:
                if not self._item_map.has_look_up_df:
                    raise SPyValueError(f'Attempted lookup for template parameter "{key}" but no lookup_df argument '
                                        f'was supplied to spy.workbooks.push(). Perhaps you are using '
                                        f'spy.workbooks.push() when you meant to use spy.push(metadata=<metadata>, '
                                        f'workbook=<workbook>)?\n{value}')

                self._override[_id] = self.look_up_id(value)



from seeq.sdk import *
from seeq.spy import _login


def is_continuation_token_used() -> bool:
    """
    Check if compute API endpoints should use the continuation_token field for pagination. On false, should use the
    old start key increment and deduplicate capsules scheme.

    Returns
    -------
    True if the SDK version is equal to or greater than 63.
    """
    return _login.is_sdk_module_version_at_least(63)


def is_force_calculated_scalars_available() -> bool:
    """
    Check if the force_calculated_scalars parameter is available on the PutScalars endpoint.

    Returns
    -------
    True if the force_calculated_scalars parameter is available on the PutScalars endpoint.
    """
    return 'force_calculated_scalars' in PutScalarsInputV1.attribute_map



from __future__ import annotations

import copy
import json
import os
import sys
import textwrap
from datetime import datetime, timedelta
from typing import Optional, Union

import pandas as pd

from seeq import spy
from seeq.base import util
from seeq.spy import _common
from seeq.spy import _login
from seeq.spy._errors import *
from seeq.spy._session import Session
from seeq.spy._status import Status
from seeq.spy.workbooks.job.data._push import load_data_results, save_data_results


@Status.handle_keyboard_interrupt()
def pull(job_folder, *, resume: bool = True, errors: Optional[str] = None, quiet: Optional[bool] = None,
         status: Optional[Status] = None, session: Optional[Session] = None) -> pd.DataFrame:
    """
    Pulls all the data that is used by the workbooks according to the Data
    Usages sections of the data_usage.json file in the job folder.

    Parameters
    ----------
    job_folder : {str}
        A full or partial path to the job folder created by
        spy.workbooks.job.pull().

    resume : bool, default True
        True if the pull should resume from where it left off, False if it
        should pull everything again.

    errors : {'raise', 'catalog'}, default 'raise'
        If 'raise', any errors encountered will cause an exception. If
        'catalog', errors will be added to a 'Result' column in the status.df
        DataFrame (errors='catalog' must be combined with
        status=<Status object>).

    quiet : bool
        If True, suppresses progress output. Note that when status is
        provided, the quiet setting of the Status object that is passed
        in takes precedence.

    status : spy.Status, optional
        If specified, the supplied Status object will be updated as the command
        progresses. It gets filled in with the same information you would see
        in Jupyter in the blue/green/red table below your code while the
        command is executed. The table itself is accessible as a DataFrame via
        the status.df property.

    session : spy.Session, optional
        If supplied, the Session object (and its Options) will be used to
        store the login session state. This is useful to log in to different
        Seeq servers at the same time or with different credentials.

    """
    _common.validate_argument_types([
        (job_folder, 'job_folder', str),
        (resume, 'resume', bool),
        (errors, 'errors', str),
        (quiet, 'quiet', bool),
        (status, 'status', Status),
        (session, 'session', Session)
    ])

    status = Status.validate(status, quiet, errors)
    session = Session.validate(session)
    _login.validate_login(session, status)

    all_usages = load_data_usage(job_folder)

    _set_up_start_end_and_calculation(all_usages)

    def _keep(_item_id, _usage):
        if not _common.present(_usage['Definition'], 'Start'):
            return False
        if not resume:
            return True
        if not util.safe_exists(get_df_filename(job_folder, _item_id)):
            return True
        return False

    pull_df = pd.DataFrame([usage['Definition'] for item_id, usage in all_usages.items() if _keep(item_id, usage)])

    data_results = load_data_results(job_folder, 'pull')
    data_results.loc[data_results['Result'] == 'Success', 'Result'] = 'Success: Already pulled'
    data_results = data_results[data_results['ID'].isin(all_usages.keys())]

    def _consumer(_row, _df):
        _df_filename = get_df_filename(job_folder, _row['ID'])
        util.safe_makedirs(os.path.dirname(_df_filename), exist_ok=True)
        _df.to_pickle(_df_filename, protocol=4)

    spy.pull(pull_df, grid=None, shape=_consumer, header='ID', status=status, session=session)

    status_df = status.df.copy()
    status_df.set_index('ID', drop=False, inplace=True)

    data_results.update(status_df, overwrite=True)
    data_results = pd.concat([data_results, status_df[~status_df.index.isin(data_results.index)]])
    save_data_results(job_folder, data_results, 'pull')

    return data_results


def get_df_filename(job_folder: str, item_id: str):
    return os.path.join(job_folder, 'Data', f'{item_id}.pickle')


def get_data_usage_filename(job_folder):
    return os.path.join(job_folder, 'data_usage.json')


def load_data_usage(job_folder):
    data_usage_filename = get_data_usage_filename(job_folder)
    with util.safe_open(data_usage_filename, 'r') as f:
        loaded = json.load(f)

    for usage_dict in loaded.values():
        for period in usage_dict['Periods']:
            period['Start'] = pd.to_datetime(period['Start'])
            period['End'] = pd.to_datetime(period['End'])

    return loaded


def save_data_usage(job_folder, data_usage_dict):
    data_usage_filename = get_data_usage_filename(job_folder)
    with util.safe_open(data_usage_filename, 'w') as f:
        return f.write(_common.safe_json_dumps(data_usage_dict))


def _set_up_start_end_and_calculation(all_usages):
    for item_id, usage_dict in all_usages.items():
        item_dict = usage_dict['Definition']
        periods = usage_dict['Periods']

        def _compare(_ts1, _ts2):
            if pd.to_datetime(_ts1) > pd.to_datetime(_ts2):
                return 1
            elif pd.to_datetime(_ts1) < pd.to_datetime(_ts2):
                return -1
            else:
                return 0

        capsules = list()
        periods = copy.deepcopy(periods)
        to_add = [period for period in periods if not period.get('Remove', False)]
        to_remove = [period for period in periods if period.get('Remove', False)]
        final_periods = list()
        for period in to_add:
            periods_to_add = [period]
            for remove_period in to_remove:
                remove_start = (remove_period['Start']
                                if remove_period['Start'] is not None else pd.Timestamp(0, tz='utc'))
                remove_end = (remove_period['End']
                              if remove_period['End'] is not None else pd.Timestamp(sys.maxsize, tz='utc'))
                for period_to_add in periods_to_add.copy():
                    if _compare(remove_start, period_to_add['End']) > 0:
                        # Period to remove is completely after subject period
                        continue
                    if _compare(remove_end, period_to_add['Start']) < 0:
                        # Period to remove is completely before subject period
                        continue

                    if (_compare(remove_start, period_to_add['Start']) > 0 >
                            _compare(remove_end, period_to_add['End'])):
                        # Period to remove bisects subject period to create two separate periods
                        periods_to_add.append({'Start': remove_end, 'End': period_to_add['End']})
                        period_to_add['End'] = remove_start
                    elif (_compare(remove_start, period_to_add['Start']) >= 0 and
                          _compare(remove_end, period_to_add['End']) >= 0):
                        # Period to remove overlaps on the right
                        period_to_add['End'] = remove_start
                    elif (_compare(remove_start, period_to_add['Start']) <= 0 and
                          _compare(remove_end, period_to_add['End']) <= 0):
                        # Period to remove overlaps on the left
                        period_to_add['Start'] = remove_end
                    elif (_compare(remove_start, period_to_add['Start']) < 0 <
                          _compare(remove_end, period_to_add['End'])):
                        # Period to remove encapsulates subject period entirely
                        periods_to_add.remove(period_to_add)

            final_periods.extend(periods_to_add)

        for period in final_periods:
            if _compare(item_dict.setdefault('Start', period['Start']), period['Start']) > 0:
                item_dict['Start'] = period['Start']
            if _compare(item_dict.setdefault('End', period['End']), period['End']) < 0:
                item_dict['End'] = period['End']

            capsules.append(f'capsule("{period["Start"].isoformat()}", "{period["End"].isoformat()}")')

        capsules_str = ',\n                   '.join(capsules)
        if 'Signal' in item_dict['Type']:
            formula = usage_dict.get('Calculation', '$signal')
            formula = formula.replace('$signal', '$final')

            item_dict['Calculation'] = textwrap.dedent(f"""
                $within = condition(
                   {capsules_str}
                )

                $final = $signal.within($within)

                {formula}
            """).strip()
        elif 'Condition' in item_dict['Type']:
            formula = usage_dict.get('Calculation', '$condition')
            formula = formula.replace('$condition', '$final')
            max_duration = item_dict.get('Maximum Duration', '40h')
            item_dict['Calculation'] = textwrap.dedent(f"""
                $touches = condition(
                  {capsules_str}
                )

                $final = $condition.removeLongerThan({max_duration}).touches($touches)

                {formula}
            """).strip()


def manifest(job_folder, *, reset=False):
    """
    Generates and returns a DataFrame with the list of items and data to be
    pulled. The manifest is initially generated by spy.workbooks.job.pull(),
    and is constructed by examining all of the Analyses and Topics that "touch"
    a signal or condition and noting the display ranges at play.

    You can modify the manifest using its sibling functions: expand(), add()
    or remove().

    The DataFrame returned is in a format suitable for spy.pull(), but in
    general it is expected that you will just use
    spy.workbooks.job.data.pull(), which has all the resume-ability of the
    spy.workbooks.job family of functions.

    Parameters
    ----------
    job_folder : {str}
        A full or partial path to the job folder created by
        spy.workbooks.job.pull().

    reset: {bool}, default False
        If True, the manifest will be reset to the original state (i.e. all
        modifications made by expand(), add(), and remove() will be
        undone).

    Returns
    -------
    pandas.DataFrame
        A DataFrame with the list of items that would be pulled by
        spy.workbooks.data.pull(). Note the presence of "Start", "End" and
        "Calculation" columns:

        Start        The earliest timestamp that will be requested for the item
        End          The latest timestamp that will be requested for the item
        Calculation  A within() or touches() calculation that will be used to
                     more precisely request the non-contiguous time periods
                     represented by the manifest.
    """
    if reset:
        _reset_manifest(job_folder, {'ID': '*'})

    all_usages = load_data_usage(job_folder)
    _set_up_start_end_and_calculation(all_usages)
    manifest_df = pd.DataFrame([row['Definition'] for item_id, row in all_usages.items()])
    manifest_df.dropna(subset=['Start'], inplace=True)
    return manifest_df


def _query(items, all_usages, job_folder):
    query_list = list()
    if not isinstance(items, list):
        query_list = [items]

    for query in query_list:
        if query is None:
            for usage_dict in all_usages.values():
                yield usage_dict
        elif isinstance(query, str):
            if query not in all_usages:
                raise SPyValueError(
                    f'Item ID {query} not found in data manifest ("{get_data_usage_filename(job_folder)}")')
            yield all_usages[query]
        elif isinstance(query, dict):
            for item_id, usage_dict in all_usages.items():
                if _common.does_definition_match_criteria(query, usage_dict['Definition']):
                    yield usage_dict
        else:
            raise SPyTypeError(f'Invalid entry in items argument: type must be str or dict, but was '
                               f'{type(query.__class__.__name__)}')


def expand(job_folder: str, items: Union[str, list, dict] = None, *, by: Union[str, timedelta, pd.Timedelta] = None,
           start_by: Union[str, timedelta, pd.Timedelta] = None, end_by: Union[str, timedelta, pd.Timedelta] = None):
    """
    Expands the start and end of the time periods to be pulled (by
    spy.workbooks.job.data.pull) by the specified amount.

    Parameters
    ----------
    job_folder : {str}
        A full or partial path to the job folder created by
        spy.workbooks.job.pull().

    items : {str, list, dict}, optional
        The items to affect in the data manifest.

        If not specified, all items are affected.

        If a string, it is interpreted as an item ID.

        If a dict, it is interpreted as a query to match
        against the properties of each item. The keys of the dict are
        property names and the values are the matching criteria, and
        can include wildcards (*, ?) and regular expressions (aka Regex).
        Regex expressions must be enclosed in forward slashes (/).

        If a list, it is interpreted as a list of item IDs and/or
        dictionaries.

    by : {str, timedelta, pd.Timedelta}, optional
        A str, timedelta or pd.Timedelta object specifying the amount of time
        to expand both the start and end times of the periods to be pulled.

    start_by : {str, timedelta, pd.Timedelta}, optional
        A str, timedelta or pd.Timedelta object specifying the amount of time
        to expand the start times of the periods to be pulled.

    end_by : {str, timedelta, pd.Timedelta}, optional
        A str, timedelta or pd.Timedelta object specifying the amount of time
        to expand the start times of the periods to be pulled.
    """
    _common.validate_argument_types([
        (job_folder, 'job_folder', str),
        (items, 'items', (str, dict, list)),
        (by, 'by', (str, timedelta, pd.Timedelta)),
        (start_by, 'start_by', (str, timedelta, pd.Timedelta)),
        (end_by, 'end_by', (str, timedelta, pd.Timedelta))
    ])

    all_usages = load_data_usage(job_folder)
    affected_item_ids = set()

    if by is not None:
        if start_by is not None or end_by is not None:
            raise SPyValueError('Cannot specify both by and start_by/end_by')
        start_by = by
        end_by = by

    if isinstance(start_by, str):
        start_by = _common.parse_str_time_to_timedelta(start_by)
    if isinstance(end_by, str):
        end_by = _common.parse_str_time_to_timedelta(end_by)

    if ((start_by is not None and start_by < pd.Timedelta(0)) or
            (end_by is not None and end_by < pd.Timedelta(0))):
        raise SPyValueError('timedeltas must be positive')

    for item in _query(items, all_usages, job_folder):
        new_periods = list()
        for period in item['Periods']:
            new_period = {'Start': period['Start'], 'End': period['End'], 'Added by User': True}
            if start_by is not None:
                new_period['Start'] = pd.to_datetime(period['Start']) - start_by
            if end_by is not None:
                new_period['End'] = pd.to_datetime(period['End']) + end_by
            new_periods.append(new_period)
        item['Periods'].extend(new_periods)
        affected_item_ids.add(item['Definition']['ID'])

    save_data_usage(job_folder, all_usages)
    return _affected_items_df(all_usages, affected_item_ids)


def add(job_folder: str, items: Union[str, list, dict] = None, *,
        start: Union[str, pd.Timestamp], end: Union[str, pd.Timestamp] = None):
    """
    Adds a time period (start and end) to the list of the time periods to be
    pulled (by spy.workbooks.job.data.pull) for a particular item or set of
    items.

    Parameters
    ----------
    job_folder : {str}
        A full or partial path to the job folder created by
        spy.workbooks.job.pull().

    items : {str, list, dict}, optional
        The items to affect in the data manifest.

        If not specified, all items are affected.

        If a string, it is interpreted as an item ID.

        If a dict, it is interpreted as a query to match
        against the properties of each item. The keys of the dict are
        property names and the values are the matching criteria, and
        can include wildcards (*, ?) and regular expressions (aka Regex).
        Regex expressions must be enclosed in forward slashes (/).

        If a list, it is interpreted as a list of item IDs and/or
        dictionaries.

    start : {str, pd.Timestamp}
        The starting time for which to pull data. This argument must be a
        string that pandas.to_datetime() can parse, or a pandas.Timestamp.

    end : {str, pd.Timestamp}, optional
        The end time for which to pull data. This argument must be a string
        that pandas.to_datetime() can parse, or a pandas.Timestamp.
        If not provided, 'end' will default to now.
    """
    _common.validate_argument_types([
        (job_folder, 'job_folder', str),
        (items, 'items', (str, dict, list)),
        (start, 'start', (str, datetime, pd.Timestamp)),
        (end, 'end', (str, datetime, pd.Timestamp)),
    ])

    all_usages = load_data_usage(job_folder)
    affected_item_ids = set()

    if start is None:
        raise SPyValueError('start argument must be specified')

    start = pd.to_datetime(start)

    if end is not None:
        end = pd.to_datetime(end)
    else:
        end = pd.Timestamp.utcnow()

    for item in _query(items, all_usages, job_folder):
        item['Periods'].append({'Start': start, 'End': end, 'Added by User': True})
        affected_item_ids.add(item['Definition']['ID'])

    save_data_usage(job_folder, all_usages)
    return _affected_items_df(all_usages, affected_item_ids)


def remove(job_folder: str, items: Union[str, list, dict] = None, *,
           start: Union[str, pd.Timestamp] = None, end: Union[str, pd.Timestamp] = None):
    """
    Removes a time period (start and end) to the set of the time periods to be
    pulled (by spy.workbooks.job.data.pull) for a particular item or set of
    items.

    Parameters
    ----------
    job_folder : {str}
        A full or partial path to the job folder created by
        spy.workbooks.job.pull().

    items : {str, list, dict}, optional
        The items to affect in the data manifest.

        If not specified, all items are affected.

        If a string, it is interpreted as an item ID.

        If a dict, it is interpreted as a query to match
        against the properties of each item. The keys of the dict are
        property names and the values are the matching criteria, and
        can include wildcards (*, ?) and regular expressions (aka Regex).
        Regex expressions must be enclosed in forward slashes (/).

        If a list, it is interpreted as a list of item IDs and/or
        dictionaries.

    start : {str, pd.Timestamp}, optional
        The starting time for the time period to remove. This argument must
        be a string that pandas.to_datetime() can parse, or a pandas.Timestamp.
        If not provided, 'start' will default to "the beginning of time,"
        meaning that all time prior to 'end' will be removed. (If neither
        'start' nor 'end' are specified, no data for the item will be pulled.)

    end : {str, pd.Timestamp}, optional
        The end time for the time period to remove. This argument must be a
        string that pandas.to_datetime() can parse, or a pandas.Timestamp.
        If not provided, 'end' will default to "the end of time," meaning that
        all time after 'start' will be removed. (If neither 'start' nor 'end'
        are specified, no data for the item will be pulled.)

    session : spy.Session, optional
        If supplied, the Session object (and its Options) will be used to
        store the login session state. This is useful to log in to different
        Seeq servers at the same time or with different credentials.
    """
    _common.validate_argument_types([
        (job_folder, 'job_folder', str),
        (items, 'items', (str, dict, list)),
        (start, 'start', (str, datetime, pd.Timestamp)),
        (end, 'end', (str, datetime, pd.Timestamp))
    ])

    all_usages = load_data_usage(job_folder)
    affected_item_ids = set()

    if start is not None:
        start = pd.to_datetime(start)

    if end is not None:
        end = pd.to_datetime(end)

    for item in _query(items, all_usages, job_folder):
        item['Periods'].append({'Start': start, 'End': end, 'Added by User': True, 'Remove': True})
        affected_item_ids.add(item['Definition']['ID'])

    save_data_usage(job_folder, all_usages)
    return _affected_items_df(all_usages, affected_item_ids)


def calculation(job_folder: str, items: Union[str, list, dict] = None, *, formula: str = None):
    """
    Apply a specific formula to the items when pulling. For example,
    you may wish to specify formula="resample($signal, 1h)" to
    reduce the density of a signal. This calculation will be applied
    for all time periods that are pulled for the item.

    Parameters
    ----------
    job_folder : {str}
        A full or partial path to the job folder created by
        spy.workbooks.job.pull().

    items : {str, list, dict}, optional
        The items to affect in the data manifest.

        If not specified, all items are affected.

        If a string, it is interpreted as an item ID.

        If a dict, it is interpreted as a query to match
        against the properties of each item. The keys of the dict are
        property names and the values are the matching criteria, and
        can include wildcards (*, ?) and regular expressions (aka Regex).
        Regex expressions must be enclosed in forward slashes (/).

        If a list, it is interpreted as a list of item IDs and/or
        dictionaries.

    formula : {str}
        The calculation to apply to the set of items. For signals, the
        formula must contain a reference to $signal, for conditions it
        must contain a reference to $condition.
    """
    _common.validate_argument_types([
        (job_folder, 'job_folder', str),
        (items, 'items', (str, dict, list)),
        (formula, 'formula', str)
    ])

    all_usages = load_data_usage(job_folder)
    affected_item_ids = set()

    if formula is None:
        raise SPyValueError('formula argument must be specified')

    for item in _query(items, all_usages, job_folder):
        for item_type in ['Signal', 'Condition']:
            if item_type in item['Definition']['Type'] and f'${item_type.lower()}' not in formula:
                raise SPyValueError(f'formula must contain a reference to f{item_type.lower()}')

        if '$final' in formula:
            raise SPyValueError('formula cannot contain $final')

        item['Calculation'] = formula
        affected_item_ids.add(item['Definition']['ID'])

    save_data_usage(job_folder, all_usages)
    return _affected_items_df(all_usages, affected_item_ids)


def _reset_manifest(job_folder: str, items: Union[str, list, dict] = None):
    """
    Resets any modifications you have made to the data manifest by
    the expand/add/remove functions.

    Parameters
    ----------
    job_folder : {str}
        A full or partial path to the job folder created by
        spy.workbooks.job.pull().

    items : {str, list, dict}, optional
        The items to affect in the data manifest.

        If not specified, all items are affected.

        If a string, it is interpreted as an item ID.

        If a dict, it is interpreted as a query to match
        against the properties of each item. The keys of the dict are
        property names and the values are the matching criteria, and
        can include wildcards (*, ?) and regular expressions (aka Regex).
        Regex expressions must be enclosed in forward slashes (/).

        If a list, it is interpreted as a list of item IDs and/or
        dictionaries.
    """
    _common.validate_argument_types([
        (job_folder, 'job_folder', str),
        (items, 'items', (str, dict, list))
    ])

    all_usages = load_data_usage(job_folder)
    affected_item_ids = set()

    for item in _query(items, all_usages, job_folder):
        new_periods = list()
        affected = False
        for period in item['Periods']:
            if period.get('Added by User', False):
                affected = True
            else:
                new_periods.append(period)

        if affected:
            item['Periods'] = new_periods
            affected_item_ids.add(item['Definition']['ID'])

        if 'Calculation' in item:
            del item['Calculation']
            affected_item_ids.add(item['Definition']['ID'])

    save_data_usage(job_folder, all_usages)
    return _affected_items_df(all_usages, affected_item_ids)


def _affected_items_df(all_usages, affected_item_ids: set):
    _set_up_start_end_and_calculation(all_usages)
    return pd.DataFrame([row['Definition'] for item_id, row in all_usages.items()
                         if item_id in affected_item_ids])


def redo(job_folder: str, status: Status):
    item_ids: pd.Series = status.df['ID']
    for index, item_id in item_ids.items():
        pickle_file = os.path.join(job_folder, 'Data', f'{item_id}.pickle')
        if util.safe_exists(pickle_file):
            util.safe_remove(pickle_file)
            status.df.at[index, 'Result'] = 'Data pull will be redone'
        else:
            status.df.at[index, 'Result'] = 'Not found'
